/* ********** Main ********** */
#connected-to-label {
  margin: 5px 15px;
  font-size: 1.2em;
  min-width: 200px;
}

#connected-status-row {
  margin-bottom: 5px;
}

#online-status-icon {
  margin-right: 10px;
}

#menu-header {
  margin: 10px 15px 10px 5px;
  font-weight: bold;
}

#menu-settings-separator {
    margin-bottom: 5px;
    font-weight: bold;
}

#current-downloads-list {
  min-height: 30px;
  min-width: 350px;
}

#current-downloads-list-placeholder {
  margin: 10px;
}

#current-downloads-list-pending-count,
#current-downloads-list-failed-count {
  margin: 0 5px;
}

#current-downloads-cover-art-image {
  margin: 3px 5px;
}

.menu-label {
    margin-right: 15px;
}

#main-menu-box {
  min-width: 230px;
}

#icon-button-box image {
  margin: 5px 2px;
  min-width: 15px;
}

#icon-button-box label {
  margin-left: 5px;
  margin-right: 3px;
}

#menu-item-download-settings,
#menu-item-clear-cache {
    min-width: 230px;
}

/* ********** Configure Provider Dialog ********** */
#ground-truth-adapter-options-list {
    margin: 0 40px;
}

#music-source-config-name-entry-grid {
    margin: 10px 0;
}

#config-verification-separator {
    margin: 5px -10px;
}

#verify-config-spinner {
    min-height: 32px;
    min-width: 32px;
}

.configure-form-help-icon {
    margin-left: 10px;
}

entry.invalid {
    border-color: red;
}

/* ********** Playlist ********** */
#playlist-list-listbox row {
  margin: 0;
  padding: 0;
}

#playlist-list-spinner:checked,
#artist-list-spinner:checked,
#drilldown-list-spinner:checked {
  margin: 10px;
  padding: 0px;
}

#playlist-list-new-playlist-entry {
  margin: 10px 10px 5px 10px;
}

#playlist-list-new-playlist-cancel {
  margin: 5px 0 10px 0;
}

#playlist-list-new-playlist-confirm {
  margin: 5px 10px 10px 0;
}

#playlist-artwork-spinner,
#artist-artwork-spinner,
#album-artwork-spinner,
#albumslist-with-songs-spinner {
  min-height: 35px;
  min-width: 35px;
}

#browse-spinner {
  min-height: 100px;
  min-width: 100px;
}

#menu-item-add-to-playlist {
    min-width: 170px;
}

#menu-item-spinner {
    margin: 10px;
}

#playlist-album-artwork {
  margin: 10px 15px 0 10px;
}

#playlist-name, #artist-detail-panel #artist-name {
  font-size: 40px;
  margin-bottom: 10px;
}

#playlist-name.collapsed,
#artist-detail-panel #artist-name.collapsed {
  font-size: 30px;
}

#playlist-comment, #playlist-stats, #artist-bio, #artist-stats, #similar-artists {
  margin-bottom: 10px;
}

#similar-artist-button {
  padding: 0;
  margin: -10px 0 0 10px;
}

/* ********** Playback Controls ********** */
#player-controls-album-artwork {
  min-height: 70px;
  min-width: 70px;
  margin-right: 10px;
}

#player-controls-bar #play-button {
  min-height: 45px;
  min-width: 35px;
  border-width: 1px;
  border-radius: 45px;
}

/* Make the play icon look centered. */
#player-controls-bar #play-button image {
  margin-left: 5px;
  margin-right: 5px;
  margin-top: 1px;
  margin-bottom: 1px;
}

#player-controls-bar #song-scrubber {
  min-width: 400px;
}

#player-controls-bar #volume-slider {
  min-width: 90px;
}

#player-controls-bar #volume-slider value {
  margin-left: 8px;
}

#player-controls-bar #song-title {
  margin-bottom: 3px;
  font-weight: bold;
}

#player-controls-bar #album-name {
  margin-bottom: 3px;
  font-style: italic;
}

#device-popover-box {
  min-width: 150px;
}

#device-type-section-title {
  margin: 5px;
  font-style: italic;
}

#up-next-popover #label {
  margin: 10px;
}

#play-queue-playing-icon {
  min-height: 40px;
  min-width: 40px;
}

#play-queue-row-image {
  min-height: 50px;
  min-width: 50px;
}

#play-queue-image-disabled {
  opacity: 0.5;
}

#play-queue-spinner {
  min-height: 35px;
  min-width: 35px;
}

/* ********** General ********** */
.menu-button {
  padding: 5px;
}

/* ********** Search ********** */
#search-results {
  min-width: 400px;
}

#search-spinner {
  margin: 15px;
}

.search-result-row {
  padding: 5px;
}

.search-result-header {
  font-weight: bold;
  font-size: 1.2em;
}

#search-artwork {
  margin-right: 5px;
  min-width: 30px;
  min-height: 30px;
}

/* ********** Error Indicator ********** */
#load-error-box {
  margin: 15px;
}

#load-error-image,
#load-error-label {
  margin-bottom: 5px;
  margin-right: 20px;
}

/* ********** Artists & Albums ********** */
#grid-artwork-spinner, #album-list-song-list-spinner {
  min-height: 35px;
  min-width: 35px;
}

#grid-artwork {
  min-height: 200px;
  min-width: 200px;
  margin: 10px;
}

#grid-spinner {
  min-height: 50px;
  min-width: 50px;
  margin: 20px;
}

#grid-header-label {
  margin-left: 10px;
  margin-right: 10px;
  margin-bottom: 3px;
  font-weight: bold;
}

#grid-info-label {
  margin-left: 10px;
  margin-right: 10px;
  margin-bottom: 10px;
}

#artist-album-artwork {
  margin: 10px 15px 0 10px;
}

#artist-album-list-artwork {
  margin: 10px;
}

#artist-album-list-album-name {
  margin: 10px 10px 5px 10px;
  font-size: 25px;
}

#album-list-song-list-spinner {
  margin: 15px;
}

@define-color box_shadow_color rgba(0, 0, 0, 0.2);

#artist-info-panel {
  box-shadow: 0 5px 5px @box_shadow_color;
  margin-bottom: 10px;
  padding-bottom: 10px;
}

#artist-detail-box {
  padding-top: 10px;
  padding-bottom: 10px;
  box-shadow: inset 0 5px 5px @box_shadow_color,
              inset 0 -5px 5px @box_shadow_color;
  background-color: @box_shadow_color;
}
